/*
 * Copyright (c) 2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: smbatvsaddr.c,v 1.4 2005/04/10 23:45:44 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/str.h"
#include "sm/mta.h"
#include "sm/rfc2821.h"
#include "sm/batv.h"
#include "sm/batvcnf.h"

#if SM_USE_BATV
/*
**  SM_BATV_SADDR -- create a signed address
**
**	Parameters:
**		address -- address to sign
**		key -- key to use
**		batv_cnf -- BATV configuration
**		batvaddr -- signed address (output)
**
**	Returns:
**		usual return code
*/

sm_ret_T
sm_batv_saddr(sm_str_P address, uint key, batv_cnf_P batv_cnf, sm_str_P batvaddr)
{
	sm_ret_T ret;
	char *addr;
	batv_T batv_sig;
	sm_a2821_T a_addr, a_part;
	sm_a2821_P pa_part;
	sm_str_P local, domain;
	time_t now;
	uint days;

	SM_REQUIRE(address != NULL);
	SM_REQUIRE(batv_cnf != NULL);
	SM_REQUIRE(batv_sig != NULL);

	now = time(NULL);

	/* simple way to calculate days since epoch, ignores leap sec etc */
	days = now / (60 * 60 * 24) + batv_cnf->batvcnf_timeout;

	local = domain = NULL;
	pa_part = NULL;
	if ((addr = (char *) sm_str_getdata(address)) == NULL)
		return sm_error_temp(SM_EM_BATV, ENOMEM);

	/* sign full address including <> */
	ret = sm_batv_sign(addr, days, key, batv_cnf, batv_sig);
	if (sm_is_err(ret))
		return ret;

	/* split address in local part and domain */
	A2821_INIT_RP(&a_addr, NULL);
	ret = t2821_scan((sm_rdstr_P) address, &a_addr, 0);
	if (sm_is_err(ret))
		goto error;
	ret = t2821_parse(&a_addr, 0);
	if (sm_is_err(ret))
		goto error;
	domain = sm_str_new(NULL, MAXADDRLEN, MAXADDRLEN);
	if (domain == NULL)
		goto enomem;
	local = sm_str_new(NULL, MAXADDRLEN, MAXADDRLEN);
	if (local == NULL)
		goto enomem;

	pa_part = &a_part;
	A2821_INIT_RP(&a_part, NULL);
	ret = t2821_extract_domain(NULL, &a_addr, &pa_part);
	if (sm_is_err(ret))
		goto error;
	ret = t2821_str(pa_part, domain);
	if (sm_is_err(ret))
		goto error;
	a2821_free(pa_part);
	pa_part = NULL;

	pa_part = &a_part;
	A2821_INIT_RP(&a_part, NULL);
	ret = t2821_extract_local(NULL, &a_addr, 0, &pa_part);
	if (sm_is_err(ret))
		goto error;
	ret = t2821_str(pa_part, local);
	if (sm_is_err(ret))
		goto error;
	a2821_free(pa_part);
	pa_part = NULL;

	ret = sm_str_printf(batvaddr, "<%S%c%x%03d%s@%S>"
			, local
			, batv_cnf->batvcnf_delim
			, key & 0x0f
			, (days % 1000)
			, batv_sig
			, domain
			);
	if (ret <= 0)
		goto error;

	SM_STR_FREE(local);
	SM_STR_FREE(domain);
	return SM_SUCCESS;

  enomem:
	ret = sm_error_temp(SM_EM_BATV, ENOMEM);
  error:
	SM_STR_FREE(local);
	SM_STR_FREE(domain);
	a2821_free(pa_part);
	return ret;
}
#endif /* SM_USE_BATV */
