#!/bin/sh
# $Id: t-alias-0.sh,v 1.15 2005/05/16 17:57:34 ca Exp $
# Copyright (c) 2003-2005 Sendmail, Inc. and its suppliers.
#	All rights reserved.
#
# By using this file, you agree to the terms and conditions set
# forth in the LICENSE file which can be found at the top level of
# the sendmail distribution.
#
# ------------------------------------------------------------
# Description:
# Script to test alias expansion for RFC2821 addresses.
# Uses smar, t-smar-1, "aliases" map, t-mm, startsmar.sh
# and t-hostname
# ------------------------------------------------------------
#
test -s stop && exit 1
#

# path to programs
P=..
SD=`dirname $0`
if test "${SD}" = "."
then
   SD=../smar
fi
. ${SD}/common.sh

SHELL=/bin/sh
OUT=alias-0.out
RES=alias-0.ip
OK=alias-0.ok
SOCK=qmsmar/qmsmar
PIDS=pids
ERRS=0
TPRG=./t-smar-1
ALI=aliases
AMAP=${ALI}.db
# Note: reference to t-hostname! XREF
HOSTNAME=`../libmta/t-hostname -l -V 2>/dev/null`
if test $? -ne 0
then
  echo "t-hostname failed"
  exit 1
fi
if test X"${HOSTNAME}" = X
then
  echo "t-hostname returned empty name"
  exit 1
fi
if echo X"${HOSTNAME}" | grep '\.' >/dev/null
then
  :
else
  echo "t-hostname returned bogus name ${HOSTNAME}"
  exit 1
fi

killit()
{
if test -s ${PIDS}
then
  # stop MTA components
  for i in `cat ${PIDS}`
  do
    kill ${i}
  done
fi
# ----------------
# thanks kids for not following the POSIX standard...
if test X"`uname`" = "XLinux"
then
  killall smar
fi
#rm mt
exit 1
}

# create map for smar
rm -f ${AMAP}
../libsmmap/t-mm <<EOF
user:	local:
user+*:	<user@local.dom>
user++:	local:
postmaster:	<user+postmaster@local.dom>
loc:	user
nodetail:	<user2@local.dom>
EOF
if test $? -ne 0
then
  echo t-mm failed
  exit 1
fi

rm -f ${SOCK} ${PIDS}
trap killit 1 2 3

# create mailertable (save old version?)
cat > mt <<EOF
local.dom	lmtp:
${HOSTNAME}	lmtp:
EOF

# start SMAR
${SD}/startsmar.sh || exit 1
rm -f ${OUT} ${RES} ${OK}

# check output somehow...
${VALGRIND} ${TPRG} -VVV '<no@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} > ${RES}
cat > ${OK} <<EOF
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

${VALGRIND} ${TPRG} -VVV '<user@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} > ${RES}
cat > ${OK} <<EOF
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

${VALGRIND} ${TPRG} -VVV '<user+det@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} > ${RES}
cat > ${OK} <<EOF
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

${VALGRIND} ${TPRG} -VVV '<postmaster@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} | sort > ${RES}
sort > ${OK} <<EOF
aliased=<user+postmaster@local.dom>
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

${VALGRIND} ${TPRG} -VVV '<user+@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} | sort > ${RES}
sort > ${OK} <<EOF
aliased=<user@local.dom>
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

${VALGRIND} ${TPRG} -M -VVV '<loc@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} | sort > ${RES}
sort > ${OK} <<EOF
aliased=<user@${HOSTNAME}>
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

${VALGRIND} ${TPRG} -VVV '<nodetail@local.dom>' > ${OUT} 2>&1
egrep '^(IP|alias)' ${OUT} | sort > ${RES}
sort > ${OK} <<EOF
aliased=<user2@local.dom>
IP[0]=${SM_LMTPIPV4D} [${SM_LMTPIPV4H}]
EOF
if diff -i ${OK} ${RES}
then
  :
else
  ERRS=`expr ${ERRS} + 1 `
  test X"${STOPONERROR}" != X && killit
fi

#${TPRG} -VVV '<ok@local.dom>' '<postmaster@local.dom>' > ${OUT} 2>&1
#egrep '^(IP|alias)' ${OUT} > ${RES}
#cat > ${OK} <<EOF
#status=0
#status=0
#EOF
#if diff -i ${OK} ${RES}
#then
#  :
#else
#  ERRS=`expr ${ERRS} + 1 `
#  test X"${STOPONERROR}" != X && killit
#fi

# stop SMAR
for i in `cat ${PIDS}`
do
  kill ${i}
done
# ----------------
# thanks kids for not following the POSIX standard...
if test X"`uname`" = "XLinux"
then
  killall smar
fi
rm mt

exit ${ERRS}
