/*
 * Copyright (c) 2002-2005 Sendmail, Inc. and its suppliers.
 *	All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 * $Id: s2q.h,v 1.37 2005/06/16 00:39:10 ca Exp $
 */

#ifndef S2Q_H
#define S2Q_H 1

#include "sm/generic.h"
#include "sm/reccom.h"
#include "statethreads/st.h"
#include "sm/rcbst.h"
#include "sm/qmgrcomm.h"
#include "smtps-str.h"
#include "sm/mta.h"

#ifndef SSQ_DEBUG
# define SSQ_DEBUG	0
#endif

#if SSQ_DEBUG
# include <stdio.h>
# define SSQ_DPRINTF(x) do {			\
	time_t currt = st_time();		\
	sm_io_fprintf(smioerr, "%6ld: ", (long) currt);	\
	sm_io_fprintf x;				\
	} while (0)

#else /* SSQ_DEBUG */
# define SSQ_DPRINTF(x)
#endif /* SSQ_DEBUG */

/* basic size for RCB in s2q_ctx */
#define S2Q_RCB_SIZE	1024

struct s2q_ctx_S
{
	ss_ctx_P	 s2q_ss_ctx;	/* pointer back to ss_ctx */
	int		 s2q_status;	/* status */
	st_netfd_t	 s2q_fd;	/* fd for communication */
	int		 s2q_smtps_id;	/* smtps id */
	st_mutex_t	 s2q_wr_mutex;	/* mutex for write */
	uint		 s2q_maxrcbs;	/* max. # of outstanding requests */
	uint		 s2q_currcbs;	/* current # of outstanding requests */

/*
**  The size of the next two arrays is equal to the number of maximum threads.
**  Therefore each thread can have one outstanding communication request.
**  The arrays are used to associate incoming RCBs with the appropriate
**  session (each thread handles one session); each RCB contains a session
**  (or transaction) id. Looking it up in s2q_sids leads to the session
**  which is stored at the same index in s2q_sess.
**  If a session can have more than one outstanding request, then the
**  the size of these arrays must be changed appropriately.
**  Note: a session uses ssse_s2q_idx to determine whether a reference to
**  it is stored in this array. If there are multiple outstanding requests,
**  then there needs to be a different way (e.g., multiple indices or some
**  other algorithm).
*/

	sessta_id_P	*s2q_sids;	/* array of session/transaction ids */
	ss_sess_P	*s2q_sess;	/* array of session ctx */
};

#define S2Q_ST_OK	0
#define S2Q_ST_IOERR	1
#define S2Q_ST_CLOSED	(-1)

/* use flags? */
#define S2Q_IS_IOERR(s2q_ctx)	((s2q_ctx)->s2q_status == S2Q_ST_IOERR)
#define S2Q_SET_IOERR(s2q_ctx)	(s2q_ctx)->s2q_status = S2Q_ST_IOERR
#define S2Q_CLR_IOERR(s2q_ctx)	(s2q_ctx)->s2q_status = S2Q_ST_OK

/* which type of server? */
#define S2Q_T_QMGR	1u
#define S2Q_T_SMAR	2u
#define S2Q_T_PMILTER	4u

sm_ret_T sm_s2q_init(s2q_ctx_P *_ps2q_ctx, ss_ctx_P _ss_ctx, const char *_sockname, uint _wait4srv, uint _max_threads, int _smtps_id, uint _maxrcbs, uint _notify);
sm_ret_T sm_s2q_stop(s2q_ctx_P _s2q_ctx);

sm_ret_T sm_s2q_nseid(ss_sess_P sess, s2q_ctx_P _s2q_ctx,
			int _smtps_id, sessta_id_P _sid);
sm_ret_T sm_s2q_cseid(ss_sess_P sess, s2q_ctx_P _s2q_ctx,
			int _smtps_id, sessta_id_P _sid);
sm_ret_T sm_s2q_ntaid(ss_sess_P sess, s2q_ctx_P _s2q_ctx,
			int _smtps_id, sessta_id_P _tid, sm_str_P _mail);
sm_ret_T sm_s2q_rcptid(ss_sess_P sess, s2q_ctx_P _s2q_ctx,
			int _smtps_id, sessta_id_P _tid, int _rcpt_idx,
			sm_str_P _rcpt);
sm_ret_T sm_s2q_ctaid(ss_sess_P sess, s2q_ctx_P _s2q_ctx,
			int _smtps_id, sessta_id_P _tid);
sm_ret_T sm_s2q_dtaid(ss_sess_P _sess, s2q_ctx_P _s2q_ctx,
			int _smtps_id, sessta_id_P _tid);
sm_ret_T sm_w4q2s_reply(ss_sess_P _sess, uint _tmo, s2q_ctx_P _s2q_ctx);

sm_ret_T sm_s2a_rcptid(ss_sess_P _ss_sess, s2q_ctx_P _s2a_ctx,
			int _smtps_id, sessta_id_P _tid, int _rcpt_idx,
			uint32_t _lt, sm_str_P _rcpt);
sm_ret_T sm_s2a_addr(ss_sess_P _ss_sess, s2q_ctx_P _s2a_ctx,
			int _smtps_id, sessta_id_P _tid,
			sm_str_P _addr, uint32_t _what, uint32_t _ltype,
			uint32_t _lflags);
sm_ret_T sm_s2a_clt(ss_sess_P _ss_sess, s2q_ctx_P _s2a_ctx,
			int _smtps_id, sessta_id_P _sid, sm_str_P _clt,
			uint32_t _what, uint32_t _ltype, uint32_t _lflags);
sm_ret_T sm_s2a_tls(ss_sess_P _ss_sess, s2q_ctx_P _s2a_ctx, int _smtps_id);

sm_ret_T ss_clr_sess_rq(s2q_ctx_P _s2q_ctx, ss_sess_P _ss_sess);

sm_ret_T ss_send_rq(ss_sess_P _ss_sess, sessta_id_P _tid, s2q_ctx_P _s2q_ctx, sm_rcb_P _rcb, bool _reply);

#endif /* S2Q_H */
